# Tcl package index file, version 1.1
# This file is generated by the "pkg_mkIndex -direct" command
# and sourced either when an application starts up or
# by a "package unknown" script.  It invokes the
# "package ifneeded" command to set up package-related
# information so that packages will be loaded automatically
# in response to "package require" commands.  When this
# script is sourced, the variable $dir must contain the
# full path name of this file's directory.

package ifneeded cfg 1.0 [list source [file join $dir cfg.tcl]]
package ifneeded log 1.1 [list source [file join $dir log.tcl]]
package ifneeded plugtcl 1.1 [list source [file join $dir plugtcl.tcl]]
package ifneeded plugtk 1.0 [list source [file join $dir plugtk.tcl]]
package ifneeded policy 1.3 [list source [file join $dir policy.tcl]]
package ifneeded rpi 1.1 [list source [file join $dir rpi.tcl]]
package ifneeded tcl::utils 1.0 [list source [file join $dir utils.tcl]]
package ifneeded tclplogo 1.0 [list source [file join $dir logo.tcl]]
package ifneeded url 1.1 [list source [file join $dir url.tcl]]
package ifneeded wait 1.1 [list source [file join $dir wait.tcl]]
